//+------------------------------------------------------------------+
//|                                                  MACD Sample.mq4 |
//|                   Copyright 2005-2014, MetaQuotes Software Corp. |
//+------------------------------------------------------------------+
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: On The Fly Self-Optimization of Expert Advisers |
//|                        https://www.mql5.com/en/blogs/post/683358 |
//+------------------------------------------------------------------+
#property copyright   "2005-2014, MetaQuotes Software Corp."
#property link        "https://www.mql5.com/en/blogs/post/683358"

#property strict

// 5-digit setup
extern int TakeProfit = 500;
input double Lots = 0.01;
extern int TrailingStop = 500;
extern double MACDOpenLevel = 30;
extern double MACDCloseLevel = 20;
extern int MATrendPeriod = 26;
input bool PrintLog = false;

#include <Optimystic.mqh>

input int HistoryDepth = 500;
input int ReoptimizationPeriod = 100;
input PERFORMANCE_ESTIMATOR Estimator = ESTIMATOR_NET_PROFIT;
input bool Enabled = false;

bool TradeEnabled = true;

class MyOptimysticCallback: public OptimysticCallback
{
  private:
    void print(const string title);

  public:
    virtual bool onBeforeOptimization();
    virtual bool onApplyParameters();
    virtual void trade(const int bar, const double Ask, const double Bid);
    virtual void onReadyParameters();
    virtual void onAfterOptimization();
};

MyOptimysticCallback myc;

Optimystic *p;

void OnInit()
{
  p = createInstance(GetPointer(myc));
  p.setEnabled(Enabled);
  p.setHistoryDepth(HistoryDepth, UNIT_BAR);
  p.setReoptimizationPeriod(ReoptimizationPeriod, UNIT_BAR);
  p.setOptimizationCriterion(Estimator);

  //p.addParameter("MACDOpenLevel", MACDOpenLevel, MACDOpenLevel * 10, MACDOpenLevel);
  //p.addParameter("MACDCloseLevel", MACDCloseLevel, MACDCloseLevel * 10, MACDCloseLevel);
  p.addParameter("MATrendPeriod", MATrendPeriod, MATrendPeriod * 10, MATrendPeriod);
  p.addParameter("TakeProfit", TakeProfit, TakeProfit * 10, TakeProfit);
  p.addParameter("TrailingStop", TrailingStop, TrailingStop * 10, TrailingStop);
}

void OnDeinit(const int reason)
{
  if(CheckPointer(p) != POINTER_INVALID) delete p;
}



void MyOptimysticCallback::trade(const int bar, const double Ask, const double Bid)
{
  double MacdCurrent, MacdPrevious;
  double SignalCurrent, SignalPrevious;
  double MaCurrent, MaPrevious;

  MacdCurrent = iMACD(NULL, 0, 12, 26, 9, PRICE_CLOSE, MODE_MAIN, bar + 0);
  MacdPrevious = iMACD(NULL, 0, 12, 26, 9, PRICE_CLOSE, MODE_MAIN, bar + 1);
  SignalCurrent = iMACD(NULL, 0, 12, 26, 9, PRICE_CLOSE, MODE_SIGNAL, bar + 0);
  SignalPrevious = iMACD(NULL, 0, 12, 26, 9, PRICE_CLOSE, MODE_SIGNAL, bar + 1);
  MaCurrent = iMA(NULL, 0, MATrendPeriod, 0, MODE_EMA, PRICE_CLOSE, bar + 0);
  MaPrevious = iMA(NULL, 0, MATrendPeriod, 0, MODE_EMA, PRICE_CLOSE, bar + 1);

  int cnt, ticket, total;

  total = OrdersTotal();
  if(total < 1 && ((getContext() != NULL) || TradeEnabled))
  {
    // no opened orders identified
    if(AccountFreeMarginCheck(Symbol(), OP_BUY, Lots) <= 0 || GetLastError() == ERR_NOT_ENOUGH_MONEY)
    {
      Print("We have no money. Free Margin = ", AccountFreeMargin());
      return;
    }
    // check for long position (BUY) possibility
    if(MacdCurrent < 0 && MacdCurrent > SignalCurrent && MacdPrevious < SignalPrevious &&
      MathAbs(MacdCurrent) > (MACDOpenLevel * Point) && MaCurrent > MaPrevious)
    {
      ticket = OrderSend(Symbol(), OP_BUY, Lots, Ask, 3, 0, Ask + TakeProfit * Point, "macd sample", 16384, 0, Green);
      if(ticket > 0)
      {
        if(OrderSelect(ticket, SELECT_BY_TICKET, MODE_TRADES))
          if(getContext() == NULL) Print("BUY order opened : ", OrderOpenPrice());
      }
      else
        Print("Error opening BUY order : ", GetLastError());
      return;
    }
    // check for short position (SELL) possibility
    if(MacdCurrent > 0 && MacdCurrent < SignalCurrent && MacdPrevious > SignalPrevious &&
      MacdCurrent > (MACDOpenLevel * Point) && MaCurrent < MaPrevious)
    {
      ticket = OrderSend(Symbol(), OP_SELL, Lots, Bid, 3, 0, Bid - TakeProfit * Point, "macd sample", 16384, 0, Red);
      if(ticket > 0)
      {
        if(OrderSelect(ticket, SELECT_BY_TICKET, MODE_TRADES))
          if(getContext() == NULL) Print("SELL order opened : ", OrderOpenPrice());
      }
      else
        Print("Error opening SELL order : ", GetLastError());
    }
    // exit from the "no opened orders" block
    return;
  }
  
  // it is important to enter the market correctly, but it is more important to exit it correctly...   
  for(cnt = 0; cnt < total; cnt++)
  {
    if(!OrderSelect(cnt, SELECT_BY_POS, MODE_TRADES))
      continue;
    if(OrderType() <= OP_SELL && // check for opened position 
      OrderSymbol() == Symbol()) // check for symbol
    {
      // long position is opened
      if(OrderType() == OP_BUY)
      {
        // should it be closed?
        if(MacdCurrent > 0 && MacdCurrent < SignalCurrent && MacdPrevious > SignalPrevious &&
          MacdCurrent > (MACDCloseLevel * Point))
        {
          // close order and exit
          if(!OrderClose(OrderTicket(), OrderLots(), Bid, 3, Violet))
            Print("OrderClose error ", GetLastError());
          return;
        }
        // check for trailing stop
        if(TrailingStop > 0)
        {
          if(Bid - OrderOpenPrice() > Point * TrailingStop)
          {
            if(OrderStopLoss() < Bid - Point * TrailingStop)
            {
              // modify order and exit
              if(!OrderModify(OrderTicket(), OrderOpenPrice(), Bid - Point * TrailingStop, OrderTakeProfit(), 0, Green))
                Print("OrderModify error ", GetLastError());
              return;
            }
          }
        }
      }
      else // go to short position
      {
        // should it be closed?
        if(MacdCurrent < 0 && MacdCurrent > SignalCurrent &&
          MacdPrevious < SignalPrevious && MathAbs(MacdCurrent) > (MACDCloseLevel * Point))
        {
          // close order and exit
          if(!OrderClose(OrderTicket(), OrderLots(), Ask, 3, Violet))
            Print("OrderClose error ", GetLastError());
          return;
        }
        // check for trailing stop
        if(TrailingStop > 0)
        {
          if((OrderOpenPrice() - Ask) > (Point * TrailingStop))
          {
            if((OrderStopLoss() > (Ask + Point * TrailingStop)) || (OrderStopLoss() == 0))
            {
              // modify order and exit
              if(!OrderModify(OrderTicket(), OrderOpenPrice(), Ask + Point * TrailingStop, OrderTakeProfit(), 0, Red))
                Print("OrderModify error ", GetLastError());
              return;
            }
          }
        }
      }
    }
  }
}

bool MyOptimysticCallback::onBeforeOptimization()
{
  p.setDeposit(::AccountBalance());
  p.setSpread((int)MarketInfo(Symbol(), MODE_SPREAD));
  return true;
}

void MyOptimysticCallback::print(const string title)
{
  double net = p.getOptimalPerformance(ESTIMATOR_NET_PROFIT);
  double pf = p.getOptimalPerformance(ESTIMATOR_PROFIT_FACTOR);
  double ddp = p.getOptimalPerformance(ESTIMATOR_DRAWDOWN_PERCENT);
  int n = (int)p.getOptimalPerformance(ESTIMATOR_TRADES_COUNT);
  if(n > 0)
  {
    Print(title, ": profit=", DoubleToString(net, 2), " PF=", (pf == DBL_MAX ? "n/a" : DoubleToString(pf, 2)), " DD%=", DoubleToString(ddp, 2), " N=", n);
  }
}

void MyOptimysticCallback::onAfterOptimization()
{
  if(p.getOptimalPerformance(Estimator) > 0)
  {
    print("Best performance");
    
    // apply new parameters for real trading
    MACDOpenLevel = p[0].getBest();
    MACDCloseLevel = p[1].getBest();
    MATrendPeriod = (int)p[2].getBest();
    Print("Best parameters: ", (string)MACDOpenLevel, " ", (string)MACDCloseLevel, " ", (string)MATrendPeriod);
    TradeEnabled = true;
  }
  else
  {
    Print("No optimum");
    TradeEnabled = false;
  }
}

bool MyOptimysticCallback::onApplyParameters()
{
  // apply parameters for backtest
  MACDOpenLevel = p[0].getValue();
  MACDCloseLevel = p[1].getValue();
  MATrendPeriod = (int)p[2].getValue();
  if(PrintLog) Print("Parameters: ", (string)MACDOpenLevel, " ", (string)MACDCloseLevel, " ", (string)MATrendPeriod);
  return true;
}

void MyOptimysticCallback::onReadyParameters()
{
  if(PrintLog) print("Performance");
}

void OnTick()
{
  if(Bars < 100 + HistoryDepth)
  {
    Print("bars less than 100 plus ", HistoryDepth);
    return;
  }

  p.onTick();
}
